import sys
import os
import io
import json
import importlib
import logging
import re
from copy import copy
from enum import Enum
from deriva.transfer.download import DerivaDownloadError, DerivaDownloadConfigurationError
from deriva.transfer.download.processors.transform.base_transform_processor import BaseTransformProcessor, \
    LOCAL_PATH_KEY, SOURCE_URL_KEY
from collections import OrderedDict

logger = logging.getLogger(__name__)


def conflate_falsish(v, falsish=''):
    if isinstance(v, (bool, int, float)) or v:
        return v
    else:
        return falsish

# this processor transforms sequence data in RBK to excel file of GEO format
# https://www.ncbi.nlm.nih.gov/geo/info/seq.html
class GeoExportTransformProcessor(BaseTransformProcessor):
    def __init__(self, envars=None, **kwargs):
        super(GeoExportTransformProcessor, self).__init__(envars, **kwargs)
        self._create_input_output_paths()

    def process(self):
        logging.info("Transforming input file [%s] into output file [%s]" %
                     (self.input_abspath, self.output_abspath))
        # input_abspath: the json file containing raw data retrieved from DB
        # output_abspath: the excel file of GEO format
        geo = Export2GEO(self.input_abspath, self.output_abspath, self.parameters, self.envars.get("hostname"))
        geo.export_all()

        super(GeoExportTransformProcessor, self).process()
        return self.outputs


# used for setting background color
class Style(Enum):
    DEFAULT = 0
    INSTRUCTION = 1
    SECTION = 2
    FIELD = 3
    HEADER = 4
    PREFACE = 5


# handling excel operations
class Export2Excel(object):
    OPENPYXL = None

    def __init__(self, output_path):
        self.__name__ = 'export_excel'
        self.wb = None
        self.ws = None
        self._output_path_ = output_path
        self.sheet_name = 'METADATA'
        self.initialize_wb()

    def import_openpyxl(self):
        # locate library
        if self.OPENPYXL is None:
            try:
                self.OPENPYXL = importlib.import_module("openpyxl")
            except ImportError as e:
                raise DerivaDownloadConfigurationError("Unable to find required module. "
                                                       "Ensure that the Python package \"openpyxl\" is installed.", e)

    def initialize_wb(self):
        self.import_openpyxl()
        # Open an xlsx
        self.wb = self.OPENPYXL.Workbook()
        self.ws = self.wb.create_sheet(index=0, title=self.sheet_name)

        # remove the default sheet
        self.wb.remove(self.wb['Sheet'])

        fgColor0 = self.OPENPYXL.styles.colors.Color(indexed=10, type='indexed')
        fgColor1 = self.OPENPYXL.styles.colors.Color(indexed=42, type='indexed')
        bgColor1 = self.OPENPYXL.styles.colors.Color(indexed=27, type='indexed')
        fgColor2 = self.OPENPYXL.styles.colors.Color(indexed=12, type='indexed')
        fgColor3 = self.OPENPYXL.styles.colors.Color(indexed=13, type='indexed')
        bgColor3 = self.OPENPYXL.styles.colors.Color(indexed=34, type='indexed')

        self.font_red = self.OPENPYXL.styles.Font(name='Arial', b=True, color=fgColor0, sz=10.0)
        self.font_blue = self.OPENPYXL.styles.Font(name='Arial', b=True, color=fgColor2, sz=10.0)
        self.font_bond = self.OPENPYXL.styles.Font(name='Arial', b=True, sz=10.0)

        self.fill_green = self.OPENPYXL.styles.PatternFill('solid', fgColor=fgColor1, bgColor=bgColor1)
        self.fill_yellow = self.OPENPYXL.styles.PatternFill('solid', fgColor=fgColor3, bgColor=bgColor3)

        self.ws.column_dimensions['A'].width = 45.0
        self.ws.column_dimensions['B'].width = 40.0
        self.ws.column_dimensions['C'].width = 24.0
        self.ws.column_dimensions['D'].width = 15.0
        for col_index in [ chr(x) for x in range( ord('A'), ord('Z') ) ]:
            self.ws.column_dimensions[col_index].width = 24.0

    @staticmethod
    # row id and column id to excel cell location
    def cell_name(row_num, col_num):
        string = ""
        while col_num > 0:
            col_num, remainder = divmod(col_num - 1, 26)
            string = chr(65 + remainder) + string
        return string + str(row_num)

    def insert_row(self, row_idx, row_amt):
        self.ws.insert_rows(idx=row_idx, amount=row_amt)

    def delete_row(self, idx, row_amt):
        self.ws.delete_rows(idx=idx, amount=row_amt)

    # writing value into cell with some style
    def write_cell(self, row_idx, col_idx, content, export_style=Style.DEFAULT):
        cell_loc = self.cell_name(row_idx, col_idx)
        cell = self.ws[cell_loc]
        cell.value = content
        if (export_style == Style.INSTRUCTION):
            cell.fill = self.fill_green
        elif (export_style == Style.HEADER):
            cell.fill = self.fill_yellow
            cell.font = self.font_blue
        elif (export_style == Style.FIELD):
            cell.font = self.font_blue
        elif (export_style == Style.SECTION):
            cell.font = self.font_red
        elif (export_style == Style.PREFACE):
            cell.fill = self.fill_green
            cell.font = self.font_bond
        else:
            pass

    def save_xlsx(self):
        self.wb.save(self._output_path_)

    def copy_cell(self, old_row_idx, old_col_idx, new_row_idx, new_col_idx, copy_style=True):
        cell = self.ws[self.cell_name(old_row_idx, old_col_idx)]
        new_cell = self.ws[self.cell_name(new_row_idx, new_col_idx)]
        new_cell.value = cell.value
        if cell.has_style and copy_style:
            new_cell.font = copy(cell.font)
            new_cell.border = copy(cell.border)
            new_cell.fill = copy(cell.fill)
            new_cell.number_format = copy(cell.number_format)
            new_cell.protection = copy(cell.protection)
            new_cell.alignment = copy(cell.alignment)
            new_cell.comment = copy(cell.comment)


class Export2GEO(object):

    def __init__(self, input_file, output_excel_path, parameters, hostname):
        self.__name__ = 'export_geo_template'
        self.input_file = input_file
        self.output_excel_path = output_excel_path
        self.parameters = parameters
        self.hostname = hostname
        self.server_env = hostname.split(".")[0]
        #logging.debug("%s %s %s %s" % (input_file, output_excel_path, parameters, hostname))
        
        self.MAX_SECTION_COLUMN_IDX = 15
        self.current_row_idx = 0
        self.header_row_idx = 0

        self.protocol_global_content_dict = {}        
        self.protocol_local = []

        self.data_processing_global_dict = {}
        self.reference_genome_set = set()
        self.data_processing_local = []
        self.INCLUDE_DATA_PROCESSING_LOCAL = False
        
        # A mapping between different protocol type to protocol columns in the Experiment table
        # There are 2 columns related to each protocol, therefore the tuple.
        self.protocol_dict = OrderedDict([
            ('growth protocol', ('Growth_Protocol', 'Growth_Protocol_Reference')),
            ('treatment protocol', ('Treatment_Protocol', 'Treatment_Protocol_Reference')),
            ('extract protocol', ('Extract_Protocol', 'Extract_Protocol_Reference')),
            ('library construction protocol', ('Construction_Protocol', 'Construction_Protocol_Reference')),
            ('isolation protocol', ('Isolation_Protocol', 'Isolation_Protocol_Reference')),
            ('label protocol', ('Label_Protocol', 'Label_Protocol_Reference')),
            ('hybridization protocol', ('Hybridization_Protocol', 'Hybridization_Protocol_Reference')),
            ('scan protocol', ('Scan_Protocol', 'Scan_Protocol_Reference')),
            ('library type', ('Library_Type', 'Library_Type_Reference')),            
        ])
        self.protocol_mandatory_list = ['growth protocol', 'treatment protocol', 'extract protocol', 'library construction protocol']

        # A list of columns in Experiment_Settings related to "Data Processing Section"
        # remove alignment since we won't export bam file. GEO prefers bam file than fastq files. If export both, they ask users to remove fastq.
        # no need for analysis metadata 
        self.data_processing_list = [
            'Data_Processing',                                     # This property is derived from Experiment
            #'Library_Type', 'Protocol_Reference',                 # library---remove since they are included under Protocol instead
            'Library_Selection', 'Strandedness', 'Used_Spike_Ins', 'Spike_Ins_Amount',   # library
            'Sequencing_Platform',  'Paired_End', 'Read_Length',   # sequencing
            #'Alignment_Format', 'Aligner', 'Aligner_Version',     # alignment---remove since we don't provide bam for now
            'Reference_Genome',
            'Quantification_Format', 'Expression_Metric', 'Quantification_Software', 'Transcriptome_Model',     #quantification
            'Visualization_Format', 'Visualization_Software', 'Visualization_Version', 'Visualization_Setting', # visuallization           
            'Duplicate_Removal', 'Sequence_Trimming', 'Pre-alignment_Sequence_Removal', 'Junction_Reads',     # analysis
            #'Typeof_Reads',  # deprecated
            'Notes'
        ]

        # direct Specimen properties that will show up in sample column headers
        # the keys will be in order when iterate        
        self.characteristics_dict = OrderedDict([
            ('Species', 'organism'),            
            ('Tissue_Names', 'source name'),
            ('Cell_Type_Names', 'characteristics: Cell Type'),
            ('Allele_Names', 'characteristics: Allele'),                        
            ('Assay_Type', 'characteristics: Assay Type'),
            ('Sex', 'characteristics: Sex'),
            ('Stage_Name', 'characteristics: Stage'),
            ('Stage_Detail', 'characteristics: Age'),
            ('Phenotype', 'characteristics: Phenotype'),
            ('Strain','characteristics: Strain'),
            ('Wild_Type', 'characteristics: Wild Type'),
            ('Passage', 'characteristics: Passage'),
            ('Upload_Notes', 'characteristics: Notes'), # change from Description to Notes
        ])

        # whether to filter study_file
        # note: ext should all be lower case as we lower case the filename before comparison
        self.study_file_pattern = ".*([.](csv|tsv|txt|mtx|bw|xlsx|h5)([.]gz)?)$"        
        # NOTE: GEO only wants fastq or bam. They prefer bam. But we do not have a way to write conditional fetch e.g. if bam, only download bam
        # For now, remove bam from the white list
        self.replicate_raw_file_pattern = ".*(r[12][.]fastq([.]gz)?)$"
        self.replicate_processed_file_pattern = ".*([.](csv|tsv|txt|mtx|bw|xlsx|h5)([.]gz)?)$"
        
        # new structure to be created
        self.experiment_dict = {}
        self.experiment_setting_dict = {}
        self.replicate_dict = {}
        self.file_dict = {}

        self.study_processed_files = []        
        self.replicate_raw_file_dict = {}
        self.replicate_processed_file_dict = {}
        self.INCLUDE_STUDY_FILES_IN_PROCESSED_FILES_SECTION = False
        
        self.initialize_data()

    def organize_experiments(self):
        # add library_strategy from experiment_setting to experiment
        for ex in self.experiment_setting_dict.values():
            self.experiment_dict[ex["Experiment_RID"]]["Library_Type"] = ex["Library_Type"]
            self.experiment_dict[ex["Experiment_RID"]]["Library_Type_Reference"] = ""

    # gathering missing fields in experiment_settings for Data Processing section
    def organize_experiment_settings(self):
        for ex in self.experiment_dict.values():
            self.experiment_setting_dict[ex["RID"]]["Data_Processing"] = conflate_falsish(ex.get("Data_Processing"), "")
        
    def organize_specimens(self):
        data = self.data
        stage = data[0].get('Stage')
        tissue = data[0].get('Tissue')
        source_type = data[0].get('Source')
        specimen_cell_type = data[0].get('Specimen_Cell_Type')
        cell_type = data[0].get('Cell_Type')
        specimen_allele = data[0].get('Specimen_Allele')
        allele = data[0].get('Allele')
        
        # -- create a list of tissue_name_set per specimen
        tissue_name_dict = {
            s["ID"]: s["Name"]  # key = value
            for s in (source_type if source_type else []) #iteration
            if s is not None  # condition
        }
        # tissue_list_dict with Specimen_RID as key
        specimen_tissue_name_set_dict = {}
        for r in (tissue if tissue else []): #iteration
            if r is None:
                continue
            specimen_tissue_name_set_dict.setdefault(r["Specimen_RID"], set()).add(tissue_name_dict[r["Tissue"]])
        # -- end

        # create a list of cell_type_name_set per specimen
        # Note: no need for cell_type_dict.
        specimen_cell_type_name_set_dict = {}
        for r in (specimen_cell_type if specimen_cell_type else []):
            if r is None:
                continue
            specimen_cell_type_name_set_dict.setdefault(r["Specimen"], set()).add(r["Cell_Type"])
            #logging.debug("-- Specimen cell type: %s %s" % (r["Specimen"], r["Cell_Type"]))
        
        # -- create a list of allele_name_set per specimen
        allele_name_dict = {
            r["RID"]: r["Name"] + ((' (' + r['Allele_Type'] + ')') if r['Allele_Type'] else '')
            for r in (allele if allele else []) #iteration
            if r is not None  # condition
        }
        # specimen_allele_name_set_dict with Specimen_RID as key
        specimen_allele_name_set_dict = {}
        for r in (specimen_allele if specimen_allele else []):
            if r is None:
                continue
            specimen_allele_name_set_dict.setdefault(r["Specimen_RID"], set()).add(allele_name_dict[r["Allele_RID"]])
        # -- end

        # stage lookup
        stage_dict = {
            r["ID"]: r
            for r in (stage if stage else []) #iteration
            if r is not None  # condition
        }

        # -- update Specimen
        for s in self.specimen_dict.values():
            specimen_rid = s["RID"]
            s["Stage_Name"] = stage_dict[s["Stage_ID"]]["Name"] if s["Stage_ID"] in stage_dict else ''
            s["Tissue_Names"] = ''.join(specimen_tissue_name_set_dict[specimen_rid]) if specimen_rid in specimen_tissue_name_set_dict else ''
            s["Cell_Type_Names"] = ''.join(specimen_cell_type_name_set_dict[specimen_rid]) if specimen_rid in specimen_cell_type_name_set_dict else ''                        
            s["Allele_Names"] = ''.join(specimen_allele_name_set_dict[specimen_rid]) if specimen_rid in specimen_allele_name_set_dict else ''
        
    ''' Organize replicate raw files and processed files
    '''
    def organize_replicate_files(self):
        # Create a file_dics lookup based on Relicate_RID
        # Assume File_Name is not null. TODO: what to do if File_Name is not null
        replicate_file_dict = {}
        for f in self.file_dict.values():
            if f is None:
                continue
            # setdefault idiom. Run help(dict.setdefault) for explanation
            replicate_file_dict.setdefault(f["Replicate_RID"], set()).add(f["RID"])
        
        # determine where files should go and update the global raw and processed file list
        for replicate_rid, files in replicate_file_dict.items():
            raw_files = []
            processed_files = []
            has_r1 = False
            has_r2 = False

            #logging.debug("begin: %s %s" % (replicate_rid, files))
            # put file into raw v.s. processed v.s. ignored
            for file_rid in files:
                file = self.file_dict[file_rid]
                filename = file["File_Name"]
                # check raw_file
                if re.match(self.replicate_raw_file_pattern, filename, re.IGNORECASE):
                    raw_files.append(file)
                elif re.match(self.replicate_processed_file_pattern, filename, re.IGNORECASE):
                    processed_files.append(file)
                else:
                    # ignore
                    pass
                    
                if re.match("r1[.]fastq([.]gz)?$", filename, re.IGNORECASE):
                    has_r1 = True
                elif re.match("r2[.]fastq([.]gz)?$", filename, re.IGNORECASE):
                    has_r2 = True
                    
            # determine paired_end based on files
            if has_r1 and has_r2:
                paired_end = 'paired-end'
            elif has_r1:
                paired_end = 'single'
            else:
                paired_end = None                
                # set the metadata from experiment_settings only if raw_files exist
                if not raw_files:
                    pass
                #logging.debug("%r" % (f,))
                experiment_rid = self.replicate_dict[replicate_rid]["Experiment_RID"]
                if  experiment_rid in self.experiment_setting_dict:
                    if self.experiment_setting_dict[experiment_rid]["Paired_End"] == "Paired End":
                        paired_end = 'paired-end'
                    elif self.experiment_setting_dict[experiment_rid]["Paired_End"] == "Single End":
                        paired_end = 'single'                        

            # update Paired_End of raw files
            for f in raw_files:
                f["Paired_End"] = paired_end
                #logging.debug("-- raw %s : %s %s" % (replicate_rid, f["File_Name"], paired_end))

            for f in processed_files:            
                logging.debug("-- processed %s : %s " % (replicate_rid, f["File_Name"]))
                
            # update global raw and processed file
            if raw_files:
                self.replicate_raw_file_dict[replicate_rid] = raw_files
            if processed_files:
                self.replicate_processed_file_dict[replicate_rid] = processed_files

        if False:
            for r, files in self.replicate_raw_file_dict.items():
                for f in files:
                    logging.debug("replicate_raw_file_dict: %s %s" % (f["File_Name"], f["Paired_End"]))
            for r, files in self.replicate_processed_file_dict.items():
                for f in files:
                    logging.debug("replicate_processed_file_dict: %s " % (f["File_Name"]))
        
    # TODO: extend DerivaDownloadError class to extend error class for custom error messages. This error cause a 502 error from the export service
    # You can set a proper message into DerivaDownloadError or your custom error, but should check if chaise displays that message when it gets the 502
    def initialize_data(self):
        logging.debug("initilaize_data: begin")
        with io.open(self.input_file, encoding='utf-8') as input_file:
            data = json.load(input_file)

        self.study = None
        # TODO: Create a proper error message            
        if not data:
            logging.debug("No Data")
            return
        
        # TODO: create a proper error message            
        if len(data) > 1:
            logging.debug("Duplicate Data")
            return

        # when there is only 1 study
        self.data = data
        # get parameters from json file
        self.study = data[0].get('Study')[0]
        self.study_files = data[0].get('Study_File')        
        self.pi = data[0].get('Principal_Investigator')[0]

        experiments = data[0].get('Experiment')
        experiment_settings = data[0].get('Experiment_Setting')
        replicates = data[0].get('Replicate')
        specimens = data[0].get('Specimen')
        files = data[0].get('File')
        
        # create a specimen_dict. Only include specimen that are sequencing type
        self.specimen_dict = {
            s["RID"]: s  # key = value
            for s in (specimens if specimens else []) #iteration
            if s is not None and re.match("(.*-Seq|NextGen)$", s["Assay_Type"], re.IGNORECASE) # condition
        }
        
        # create a replicate_dict. Using succint syntax.
        # Include only sequencing replicates
        self.replicate_dict = {
            r["RID"]: r  # key = value
            for r in (replicates if replicates else []) #iteration
            if r is not None and (r["Specimen_RID"] in self.specimen_dict)  # condition
        }

        exp_with_seq_replicates = {r["Experiment_RID"] for r in self.replicate_dict.values() }
        
        # -- create a dictionary for looking up different data
        # create a experiment_dict that is of Seq type only
        self.experiment_dict = {
            e["RID"]: e  # key = value
            for e in (experiments if experiments else []) #iteration
            if e is not None and e["RID"] in exp_with_seq_replicates # condition
        }
        
        #
        # TODO: if there is no seq experiment, throw an error
        #
        
        # create a experiment_settings_dict. Using succint syntax
        self.experiment_setting_dict = {
            e["Experiment_RID"]: e  # key = value
            for e in (experiment_settings if experiment_settings else []) #iteration
            if e is not None and (e["Experiment_RID"] in self.experiment_dict) # condition
        }

        # create a file_dict. Using succint syntax
        self.file_dict = {
            f["RID"]: f  # key = value
            for f in (files if files else []) #iteration
            if f is not None and (f["Replicate_RID"] in self.replicate_dict)  # condition
        }

        # -- update self.study_files, self.replicate_raw_files, self.replicate_processed_files
        # create self.study_files
        for f in (self.study_files if self.study_files else []):
            if f is None:
                continue
            if re.match(self.study_file_pattern, f["File_Name"], re.IGNORECASE):
                self.study_processed_files.append(f)
        
        # sort the raw and processed files
        self.study_processed_files.sort( key = lambda x: (x.get('File_Name')))
        # -- debug
        #for f in self.study_processed_files:
        #    logging.debug("study: %s " % (f["File_Name"]))                

        # -- organizing data from different tables in the appropriate structure so data in different sections can be retrieved uniformly
        self.organize_experiments()
        self.organize_experiment_settings()                
        self.organize_specimens()
        self.organize_replicate_files()

        #
        # TODO: Add validation to check that all replicates have at least 1 raw files
        #
        
        #print("initilaize_data: end")


    def export_start(self):
        self.excel = Export2Excel(self.output_excel_path)

    def export_all(self):
        if self.study is not None:
            # export data for different sections such as series,samples etc.
            self.export_start()
            self.export_preface()
            self.export_series()
            self.export_sample_prepare()
            self.export_sample()
            self.export_protocol()
            self.export_data_processing()
            self.export_processed_datafiles()
            self.export_raw_datafiles()
            self.export_paired_end()
            self.export_finish()

    # export preface sections in GEO template excel
    # words may need update if GEO template update
    # https://www.ncbi.nlm.nih.gov/geo/info/seq.html#metadata
    def export_preface(self):
        #print("export_preface: start")
        self.current_row_idx = 1
        self.excel.write_cell(self.current_row_idx, 1, '# High-throughput sequencing metadata template (version 2.1).',
                              Style.PREFACE)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, '# All fields in this template must be completed.',
                              Style.PREFACE)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1,
                              '# Templates containing example data are found in the METADATA EXAMPLES spreadsheet tabs at the foot of this page.',
                              Style.PREFACE)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1,
                              '# Field names (in blue on this page) should not be edited. Hover over cells containing field names to view field content guidelines.',
                              Style.PREFACE)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, (
            '# Human data. If there are patient privacy concerns regarding making data fully public through GEO, '
            'please submit to NCBI''s dbGaP (http://www.ncbi.nlm.nih.gov/gap/) database. dbGaP has controlled access'
            'mechanisms and is an appropriate resource for hosting sensitive patient data.'), Style.PREFACE)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, '', Style.PREFACE)
        self.current_row_idx += 1

        for col in range(2, self.MAX_SECTION_COLUMN_IDX):
            for row in range(1, 7):
                self.excel.write_cell(row, col, '', Style.PREFACE)
        #print("export_preface: end")

    # export SERIES(study)
    def export_series(self):
        ##print("export_series: begin")
        s = self.study
        self.excel.write_cell(self.current_row_idx, 1, 'SERIES', Style.SECTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, '# This section describes the overall experiment.',
                              Style.INSTRUCTION)
        for col in range(2, self.MAX_SECTION_COLUMN_IDX):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'title', Style.FIELD)
        self.excel.write_cell(self.current_row_idx, 2, self.study.get('Title', ''))
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'summary', Style.FIELD)
        self.excel.write_cell(self.current_row_idx, 2, self.study.get('Summary', ''))
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'overall design', Style.FIELD)
        self.excel.write_cell(self.current_row_idx, 2, self.study.get('Overall_Design', ''))
        self.current_row_idx += 1

        consortium_study_id = ''
        if s["Consortium"] == 'GUDMAP':
            consortium = 'GUDMAP Consortium'            
            consortium_study_id = ''.join(['https://', self.server_env, '.', 'gudmap.org/id/', s["RID"]])
        elif s["Consortium"] == 'RBK':
            consortium = '(Re)Building A Kidney (RBK) Consortium'                        
            consortium_study_id = ''.join(['https://', self.server_env, '.', 'rebuildingakidney.org/id/', s["RID"]])
        if consortium_study_id:
            study_source = "Source: " + consortium + ", Study Record ID: " + consortium_study_id
            # remove blank space
            #self.excel.write_cell(self.current_row_idx, 1, 'overall design', Style.FIELD)
            #self.excel.write_cell(self.current_row_idx, 2, '')
            #self.current_row_idx += 1
            self.excel.write_cell(self.current_row_idx, 1, 'overall design', Style.FIELD)
            self.excel.write_cell(self.current_row_idx, 2, study_source)
            self.current_row_idx += 1
        
        self.excel.write_cell(self.current_row_idx, 1, 'contributor', Style.FIELD)
        self.excel.write_cell(self.current_row_idx, 2, self.pi.get('Full_Name', ''))
        self.current_row_idx += 1

        # study-level processed file: one line for each study file
        for pf in self.study_processed_files:
            file_path = self.parameters.get("study_files_path_template", pf.get("File_Name", ""))
            self.excel.write_cell(self.current_row_idx, 1, 'supplementary file', Style.FIELD)
            self.excel.write_cell(self.current_row_idx, 2, file_path.format(**pf))
            self.current_row_idx += 1

        # write an empty line (to be over-written if the self.study_processed_file is empty        
        if not self.study_processed_files:
            self.excel.write_cell(self.current_row_idx, 1, 'supplementary file', Style.FIELD)        
            self.excel.write_cell(self.current_row_idx, 2, '')
            self.current_row_idx += 1
            
        #print("export_series: end")
        
    # prepare the data for sample export
    def export_sample_prepare(self):
        #print("export_sample_prepare: begin")

        # -- protocol
        # Check shared protocols: protocols that are the same for all experiments with seq replicates
        # Protocols which are applicable to only a subset of Samples should be included
        # as additional columns of the SAMPLES section instead.
        
        # protocol_global_dict stores items that are globally unique
        self.protocol_local = []
        self.protocol_global_dict = {}
        
        for p, col_list in self.protocol_dict.items():
            protocol_set = set()
            for e in self.experiment_dict.values():
                #print("%s: %s, %s : %s, %s" % (e["RID"], col_list[0], col_list[1], e.get(col_list[0]), e.get(col_list[1])))
                protocol_tuple = tuple( conflate_falsish(e.get(col), '') for col in col_list )
                protocol_set.add(protocol_tuple)
                if len(protocol_set) > 1:
                    break
            protocol_len = len(protocol_set)
            if protocol_len == 1:
                protocol_tuple = protocol_set.pop()
                if ''.join(protocol_tuple):
                    self.protocol_global_dict[p] = protocol_tuple
            elif protocol_len > 1:                 # there is content                
                self.protocol_local.append(p)
                
        logging.debug("-- global protocol : %s" % (self.protocol_global_dict.keys()))
        logging.debug("++ local protocol: %s" % (self.protocol_local))

        # -- data processing steps
        # Check shared data processing: experiment settings that are the same for all experiments with seq replicates
        # Properties which are applicable to only a subset of Samples should be included
        # as additional columns of the SAMPLES section instead.

        self.data_processing_local = []
        self.data_processing_global_dict = {}
        
        for dp in self.data_processing_list:
            data_processing_set = set()
            for es in self.experiment_setting_dict.values():
                content = str(conflate_falsish(es.get(dp), ''))
                data_processing_set.add(content)
                if len(data_processing_set) > 1:
                    #logging.debug("!!dp:%s --> Experiment_RID:%s content:%s set:%s " % (dp, es["Experiment_RID"], content, data_processing_set))
                    break
            data_processing_len = len(data_processing_set)
            if data_processing_len == 1:
                if content:
                    self.data_processing_global_dict[dp] = content
            elif data_processing_len > 1:
                self.data_processing_local.append(dp)
                
        logging.debug("-- global data_processing : %s" % (self.data_processing_global_dict.keys()))
        logging.debug("++ local data_processing: %s" % (self.data_processing_local))

        # -- reference genome under Data Processing section
        # Build Reference Genome Set for including in the Data Processing section
        self.reference_genome_set = set()        
        for e in self.experiment_setting_dict.values():
            content = conflate_falsish(e.get("Reference_Genome"), '')
            if content:            
                self.reference_genome_set.add(content)
                

    def format_protocol_tuple(self, protocol_tuple):
        content =  protocol_tuple[0] + (("\nReference: " + protocol_tuple[1]) if protocol_tuple[1] else '')        
        #logging.debug("format_protocol_tuple: %s --> %s" % (protocol_tuple, content))
        return content

    
    def print_experiment_protocol(self, experiment, protocol):
        cols = self.protocol_dict[protocol]
        col1 = conflate_falsish(experiment.get(cols[0]), '')
        col2 = conflate_falsish(experiment.get(cols[1]), '')
        content =  col1 + (("\nReference: " + col2) if col2 else '')
        return content
        
        
    # export data of all the samples in one study
    def export_sample(self):
        #print("export_sample: begin")
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'SAMPLES', Style.SECTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1,
                              ('# This section lists and describes each of the biological Samples under investgation,'
                               'as well as any protocols that are specific to individual Samples.'), Style.INSTRUCTION)
        # write some blank cells to make words of last cell showing expanded.
        for col in range(2, self.MAX_SECTION_COLUMN_IDX):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1,
                              '# Additional "processed data file" or "raw file" columns may be included.',
                              Style.INSTRUCTION)
        # write some blank cells to make words of last cell showing expanded.
        for col in range(2, self.MAX_SECTION_COLUMN_IDX):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.header_row_idx = self.current_row_idx + 1
        self.current_row_idx += 1

        # check the existance of specimen direct properties for all specimens. These will form the column headings
        null_characteristics = set(self.characteristics_dict.keys())
        for s in self.specimen_dict.values():
            if not null_characteristics:
                break
            for c in list(null_characteristics):
                if s.get(c):
                    null_characteristics.remove(c) 
        
        characteristics_exist = []
        for c in self.characteristics_dict.keys():
            if c not in null_characteristics:            
                characteristics_exist.append(c) 

        # == write SAMPLES
        # experiment by experiment
        for r in self.replicate_dict.values():
            specimen_rid = r["Specimen_RID"]
            if specimen_rid in self.specimen_dict:
                specimen = self.specimen_dict[specimen_rid]
            else:
                continue
            
            if r['Experiment_RID'] in self.experiment_dict:
                e = self.experiment_dict[r['Experiment_RID']]
                bio_replicate = str(r.get("Biological_Replicate_Number"))
                tech_replicate = str(r.get("Technical_Replicate_Number"))
                # construct sample_name using internal_ID,Biological_Replicate_Number and Technical_Replicate_Number
                sample_name = ''.join([ e.get('Internal_ID', ''), '_', bio_replicate, '_', tech_replicate])
                sample_title = ''.join([
                    e.get('Name', ''),
                    ' (biological replicate ', bio_replicate,
                    ', technical replicate ', tech_replicate,
                    ')'
                ])
                sample_molecule = e.get('Molecule_Type', '')

                
            # maximum number of replicate raw files
            max_num_raw_files = 0
            for v in self.replicate_raw_file_dict.values():
                num_files = len(v)
                if num_files > max_num_raw_files:
                    max_num_raw_files = num_files 

            # -- start writing
            local_col_idx = 1
            self.current_row_idx += 1

            # -- properties from experiment
            self.excel.write_cell(self.header_row_idx, local_col_idx, 'sample name', Style.HEADER)
            self.excel.write_cell(self.current_row_idx, local_col_idx, sample_name)
            local_col_idx += 1
            self.excel.write_cell(self.header_row_idx, local_col_idx, 'title', Style.HEADER)
            self.excel.write_cell(self.current_row_idx, local_col_idx, sample_title)
            local_col_idx += 1
            self.excel.write_cell(self.header_row_idx, local_col_idx, 'molecule', Style.HEADER)
            self.excel.write_cell(self.current_row_idx, local_col_idx, sample_molecule)
            local_col_idx += 1
            
            # -- Specimen properties
            if r["Specimen_RID"] in self.specimen_dict:
                s = self.specimen_dict[r["Specimen_RID"]]
                for c in characteristics_exist:
                    characteristic = s.get(c, '')
                    self.excel.write_cell(self.header_row_idx, local_col_idx, self.characteristics_dict[c], Style.HEADER)
                    self.excel.write_cell(self.current_row_idx, local_col_idx, characteristic)
                    local_col_idx += 1

            # -- Reference to replicate ID
            column_name = ''.join(["characteristics: ", r["Consortium"], " Record ID"])
            consortium_id = ''
            if r["Consortium"] == 'GUDMAP':
                consortium_id = ''.join(['https://', self.server_env, '.', 'gudmap.org/id/', r["RID"]])
            elif r["Consortium"] == 'RBK':
                consortium_id = ''.join(['https://', self.server_env, '.', 'rebuildingakidney.org/id/', r["RID"]])
            if consortium_id:
                self.excel.write_cell(self.header_row_idx, local_col_idx, column_name, Style.HEADER)
                self.excel.write_cell(self.current_row_idx, local_col_idx, consortium_id)
                local_col_idx += 1

            # -- add local data processing here under characteristics
            if self.INCLUDE_DATA_PROCESSING_LOCAL and (r['Experiment_RID'] in self.experiment_setting_dict):
                es = self.experiment_setting_dict[r['Experiment_RID']]
                for dp in self.data_processing_local:
                    column_name = "characteristics: " + dp
                    content = str(conflate_falsish(es.get(dp), ''))
                    self.excel.write_cell(self.header_row_idx, local_col_idx, column_name, Style.HEADER)
                    self.excel.write_cell(self.current_row_idx, local_col_idx, content)
                    local_col_idx += 1
            
            # -- local protocol
            # if different contents for protocol_type, need to add column for this protocol_type
            # then different samples may has different content to write
            for p in self.protocol_local:
                col_list = self.protocol_dict[p]
                protocol_tuple = tuple( conflate_falsish(e.get(col), '') for col in col_list )
                content = self.format_protocol_tuple(protocol_tuple)
                self.excel.write_cell(self.header_row_idx, local_col_idx, p, Style.HEADER)
                self.excel.write_cell(self.current_row_idx, local_col_idx, content)
                local_col_idx += 1

            # checkpoint before file listing, so begging of processed_data_file column can be determined
            processed_file_col_idx = local_col_idx + max_num_raw_files

            # -- raw files
            if r["RID"] in self.replicate_raw_file_dict:
                for f in self.replicate_raw_file_dict[r["RID"]]:
                    file_path = self.parameters.get("replicate_files_path_template", f.get("File_Name", ""))
                    self.excel.write_cell(self.header_row_idx, local_col_idx, 'raw file', Style.HEADER)
                    self.excel.write_cell(self.current_row_idx, local_col_idx, file_path.format(**f))
                    local_col_idx += 1

            # -- processed files
            # replicate processed files. Move col to end of raw files
            local_col_idx = processed_file_col_idx
            if r["RID"] in self.replicate_processed_file_dict:
                for f in self.replicate_processed_file_dict[r["RID"]]:
                    file_path = self.parameters.get("replicate_files_path_template", f.get("File_Name", ""))
                    self.excel.write_cell(self.header_row_idx, local_col_idx, 'processed data file', Style.HEADER)
                    self.excel.write_cell(self.current_row_idx, local_col_idx, file_path.format(**f))
                    local_col_idx += 1
            else:
                self.excel.write_cell(self.header_row_idx, local_col_idx, 'processed data file', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, local_col_idx, 'none')
                local_col_idx += 1
        
        #print("export_sample: end")
    
    # export data for protocol section
    def export_protocol(self):
        #print("export_protocol: begin")
        self.current_row_idx = self.current_row_idx + 2
        self.excel.write_cell(self.current_row_idx, 1, 'PROTOCOLS', Style.SECTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1,
                              ('# Any of the protocols below which are applicable to only a subset of '
                               'Samples should be included as additional columns of the SAMPLES section instead.'),
                              Style.INSTRUCTION)
        # write some blank cells to make words of last cell showing expanded.
        for col in range(2, self.MAX_SECTION_COLUMN_IDX):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1

        # print the mandatory protocol first, regardless of content
        for p in self.protocol_mandatory_list:
            if p in self.protocol_global_dict:
                content = self.format_protocol_tuple(self.protocol_global_dict[p])
            else:
                content = ''
            #logging.debug(" -- global protocol (required) %s : %s --" % (p, content))
            self.excel.write_cell(self.current_row_idx, 1, p, Style.FIELD)
            self.excel.write_cell(self.current_row_idx, 2, content)
            self.current_row_idx += 1

        # print the rest only if there is content
        for p in self.protocol_global_dict:
            if p in self.protocol_mandatory_list:
                continue
            content = self.format_protocol_tuple(self.protocol_global_dict[p])
            #logging.debug(" -- global protocol (optional) %s : %s --" % (p, content))            
            self.excel.write_cell(self.current_row_idx, 1, p, Style.FIELD)
            self.excel.write_cell(self.current_row_idx, 2, content)            
            self.current_row_idx += 1

        #print("export_protocol: end")
    
    # export data processing pipline information
    def export_data_processing(self):
        #print("export_data_processing: begin")
        # DATA PROCESSING PIPELINE
        self.current_row_idx = self.current_row_idx + 2

        self.excel.write_cell(self.current_row_idx, 1, 'DATA PROCESSING PIPELINE', Style.SECTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1,
                              ('# Data processing steps include base-calling, alignment, filtering,'
                               'peak-calling, generation of normalized abundance measurements etc'), Style.INSTRUCTION)
        for col in range(2, self.MAX_SECTION_COLUMN_IDX):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1,
                              ('# For each step provide a description, as well as software name,'
                               'version, parameters, if applicable.'), Style.INSTRUCTION)
        for col in range(2, self.MAX_SECTION_COLUMN_IDX):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, '# Include additional steps, as necessary.', Style.INSTRUCTION)
        for col in range(2, self.MAX_SECTION_COLUMN_IDX):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1

        # print data processing steps
        for dp in self.data_processing_global_dict:
            content = dp + " : " + self.data_processing_global_dict[dp]
            #logging.debug(" ++ global data processing %s : %s --" % (dp, content))            
            self.excel.write_cell(self.current_row_idx, 1, "data processing step", Style.FIELD)
            self.excel.write_cell(self.current_row_idx, 2, content)            
            self.current_row_idx += 1

        # print genome build
        if self.reference_genome_set:
            content = ', '.join(self.reference_genome_set)
            self.excel.write_cell(self.current_row_idx, 1, 'genome build', Style.FIELD)
            self.excel.write_cell(self.current_row_idx, 2, content)
            self.current_row_idx += 1
        
        self.excel.write_cell(self.current_row_idx, 1, 'processed data files format and content', Style.FIELD)
        self.excel.write_cell(self.current_row_idx, 2, 'processed data file format')
        self.current_row_idx += 1

        #print("export_data_processing: end")
    
    # export processed datafile names of this study
    def export_processed_datafiles(self):
        #print("export_processed_datafiles: begin")        
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1,
                              ('# For each file listed in the "processed data file" columns of '
                               'the SAMPLES section, provide additional information below.'), Style.INSTRUCTION)
        for col in range(2, self.MAX_SECTION_COLUMN_IDX):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'PROCESSED DATA FILES', Style.SECTION)
        self.header_row_idx = self.current_row_idx + 1
        self.current_row_idx += 2

        # -- replicate level processed files
        for replicate_rid in sorted(self.replicate_processed_file_dict):
            for pf in self.replicate_processed_file_dict[replicate_rid]:  #.sort(key = lambda x: x.get('File_Name')):
                local_col_idx = 1
                file_path = self.parameters.get("replicate_files_path_template", pf.get("File_Name", ""))
                file_path = file_path.format(**pf)
                self.excel.write_cell(self.header_row_idx, local_col_idx, 'file name', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, local_col_idx, file_path)
                local_col_idx += 1
                self.excel.write_cell(self.header_row_idx, local_col_idx, 'file type', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, local_col_idx, pf.get('File_Type', ''))
                local_col_idx += 1
                self.excel.write_cell(self.header_row_idx, local_col_idx, 'file checksum', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, local_col_idx, pf.get('MD5', ''))
                self.current_row_idx += 1

        # -- study level file. GEO doesn't want study-level processed files in this section
        # -- leave this false for now.
        if self.INCLUDE_STUDY_FILES_IN_PROCESSED_FILES_SECTION:
            for pf in self.study_processed_files:
                local_col_idx = 1
                file_path = self.parameters.get("study_files_path_template", pf.get("File_Name", ""))
                file_path = file_path.format(**pf)
                self.excel.write_cell(self.header_row_idx, local_col_idx, 'file name', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, local_col_idx, file_path)
                local_col_idx += 1
                self.excel.write_cell(self.header_row_idx, local_col_idx, 'file type', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, local_col_idx, pf.get('File_Type', ''))
                local_col_idx += 1
                self.excel.write_cell(self.header_row_idx, local_col_idx, 'file checksum', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, local_col_idx, pf.get('MD5', ''))
                self.current_row_idx += 1
                #print("export_processed_datafiles: end")        
        
    # export raw datafile names of this study
    def export_raw_datafiles(self):
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, ('# For each file listed in the "raw file" columns of'
                                                        'the SAMPLES section, provide additional information below.'),
                              Style.INSTRUCTION)
        for col in range(2, self.MAX_SECTION_COLUMN_IDX):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'RAW FILES', Style.SECTION)
        self.header_row_idx = self.current_row_idx + 1
        self.current_row_idx += 2

        for replicate_rid in sorted(self.replicate_raw_file_dict):
            es = self.experiment_setting_dict[self.replicate_dict[replicate_rid]["Experiment_RID"]]
            for pf in sorted(self.replicate_raw_file_dict[replicate_rid], key = lambda x: x.get('File_Name')):
                current_col_idx = 1
                file_path = self.parameters.get("replicate_files_path_template", pf.get("File_Name", ""))
                self.excel.write_cell(self.header_row_idx, current_col_idx, 'file name', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, current_col_idx, file_path.format(**pf))
                current_col_idx += 1
                self.excel.write_cell(self.header_row_idx, current_col_idx, 'file type', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, current_col_idx, pf.get('File_Type', ''))
                current_col_idx += 1
                self.excel.write_cell(self.header_row_idx, current_col_idx, 'file checksum', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, current_col_idx, pf.get('MD5', ''))
                current_col_idx += 1
                self.excel.write_cell(self.header_row_idx, current_col_idx, 'instrument model', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, current_col_idx, es.get('Sequencing_Platform', ''))
                current_col_idx += 1
                self.excel.write_cell(self.header_row_idx, current_col_idx, 'single or paired-end', Style.HEADER)
                self.excel.write_cell(self.current_row_idx, current_col_idx, pf.get('Paired_End', ''))
                current_col_idx += 1
                self.current_row_idx += 1
    
    def export_paired_end(self):
        #print("export_paired_end: begin")        
        # PAIRED-END EXPERIMENTS
        # todo need check what's paired-end data look like
        # currently, we haven't seen paired-end data
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, (
            '# For paired-end experiments, list the 2 associated raw files, and provide average insert size and standard deviation,'
            'if known. For SOLiD experiments, list the 4 file names (include "file name 3" and "file name 4" columns).'),
                              Style.INSTRUCTION)
        for col in range(2, self.MAX_SECTION_COLUMN_IDX):
            self.excel.write_cell(self.current_row_idx, col, '', Style.INSTRUCTION)
        self.current_row_idx += 1
        self.excel.write_cell(self.current_row_idx, 1, 'PAIRED-END EXPERIMENTS', Style.SECTION)
        self.header_row_idx = self.current_row_idx + 1
        self.current_row_idx += 1
        current_col_idx = 1
        self.excel.write_cell(self.header_row_idx, current_col_idx, 'file name 1', Style.HEADER)
        current_col_idx += 1
        self.excel.write_cell(self.header_row_idx, current_col_idx, 'file name 2', Style.HEADER)
        current_col_idx += 1
        self.excel.write_cell(self.header_row_idx, current_col_idx, 'average insert size', Style.HEADER)
        current_col_idx += 1
        self.excel.write_cell(self.header_row_idx, current_col_idx, 'standard deviation', Style.HEADER)
        #print("export_paired_end: end")

    
    def export_finish(self):
        self.excel.save_xlsx()



import os
import sys
import json
import traceback
from deriva.transfer import DerivaUpload, DerivaUploadError, DerivaUploadConfigurationError, \
    DerivaUploadCatalogCreateError, DerivaUploadCatalogUpdateError
from deriva.core import BaseCLI, write_config, format_credential, format_exception, urlparse


class DerivaUploadCLI(BaseCLI):
    def __init__(self, uploader, description, epilog):
        if not issubclass(uploader, DerivaUpload):
            raise TypeError("DerivaUpload subclass required")

        BaseCLI.__init__(self, description, epilog, uploader.getVersion(), hostname_required=True)
        self.parser.add_argument('--no-config-update', action="store_true",
                                 help="Do not check for (and download) an updated configuration from the server.")
        self.parser.add_argument('--purge-state', action="store_true",
                                 help="Purge (delete) any existing transfer state files found in the directory "
                                      "hierarchy of the input path.")
        self.parser.add_argument('--dry-run', action="store_true",
                                 help="Scan the input directory for matching files to upload, "
                                      "but do not transfer any files.")
        self.parser.add_argument('--output-file', metavar='<file>',
                                 help="Optional path where a JSON-formatted output file will be written, "
                                      "containing file upload status and associated metadata.")
        self.parser.add_argument("--catalog", default=1, metavar="<1>", help="Catalog number. Default: 1")
        self.parser.add_argument("path", metavar="<input dir>", help="Path to an input directory.")
        self.uploader = uploader

    @staticmethod
    def upload(uploader,
               data_path,
               hostname,
               catalog=1,
               token=None,
               config_file=None,
               credential_file=None,
               no_update=False,
               purge=False,
               dry_run=False,
               output_file=None):

        if not issubclass(uploader, DerivaUpload):
            raise TypeError("DerivaUpload subclass required")

        assert hostname
        server = dict()
        server["catalog_id"] = catalog
        if hostname.startswith("http"):
            url = urlparse(hostname)
            server["protocol"] = url.scheme
            server["host"] = url.netloc
        else:
            server["protocol"] = "https"
            server["host"] = hostname

        deriva_uploader = uploader(config_file, credential_file, server, dcctx_cid="cli/" + DerivaUploadCLI.__name__)
        if token:
            deriva_uploader.setCredentials(format_credential(token))
        if not config_file and not no_update:
            config = deriva_uploader.getUpdatedConfig()
            if config:
                write_config(deriva_uploader.getDeployedConfigFilePath(), config)
        if not deriva_uploader.isVersionCompatible():
            raise RuntimeError("Version incompatibility detected", "Current version: [%s], required version(s): %s." % (
                deriva_uploader.getVersion(), deriva_uploader.getVersionCompatibility()))
        deriva_uploader.scanDirectory(data_path, abort_on_invalid_input=False, purge_state=purge)
        if not dry_run:
            results = deriva_uploader.uploadFiles()
            if output_file:
                with open(output_file, "w") as output:
                    json.dump(results, output)
        deriva_uploader.cleanup()

    def main(self):
        args = self.parse_cli()
        if args.path is None:
            sys.stderr.write("\nError: Input directory not specified.\n")
            self.parser.print_usage()
            return 2
        if not args.quiet:
            sys.stderr.write("\n")

        try:
            DerivaUploadCLI.upload(self.uploader,
                                   os.path.abspath(args.path),
                                   args.host,
                                   args.catalog,
                                   args.token,
                                   args.config_file,
                                   args.credential_file,
                                   args.no_config_update,
                                   args.purge_state,
                                   args.dry_run,
                                   args.output_file)
        except (RuntimeError, FileNotFoundError, DerivaUploadError, DerivaUploadConfigurationError,
                DerivaUploadCatalogCreateError, DerivaUploadCatalogUpdateError) as e:
            sys.stderr.write(("\n" if not args.quiet else "") + format_exception(e))
            if args.debug:
                traceback.print_exc()
            return 1
        except:
            traceback.print_exc()
            return 1
        finally:
            if not args.quiet:
                sys.stderr.write("\n\n")
        return 0



import sys
import json
import re
from deriva.core import ErmrestCatalog, AttrDict, ermrest_model, get_credential
from deriva.config.base_config import BaseSpec, BaseSpecList, ConfigUtil, ConfigBaseCLI

MY_VERSION = 0.99


class NoForeignKeyError(ValueError):
    pass


class AttrSpecList(BaseSpecList):
    SPEC_TYPES = ["catalog_annotations", "schema_annotations", "table_annotations", "column_annotations",
                  "foreign_key_annotations"]

    def __init__(self, known_attrs, specdict, strict=False):
        self.ignore_unmanaged = False
        self.managed_annotations = self.annotation_list(known_attrs.get(u'managed'))
        if self.managed_annotations is None:
            raise ValueError("No 'managed' attribute list")
        if known_attrs.get(u'ignore_all_unmanaged'):
            self.ignore_unmanaged = True
        self.ignored_annotations = self.annotation_list(known_attrs.get(u'ignored'))
        if self.ignored_annotations is None:
            self.ignored_annotations = []
        # dictlist = dictlist + [{"managed_annotations": self.managed_annotations}, {"ignored_annotations": self.ignored_annotations}, {"ignore_all_unmanaged": self.ignore_unmanaged}]
        BaseSpecList.__init__(self, AttrSpec, specdict, strict)

    def annotation_list(self, orig_list):
        if orig_list is None:
            return None
        new = []
        for item in orig_list:
            new.append(unicode(item))
        return new

    def add_list(self, dictlist):
        for d in dictlist:
            if len(d) > 0:
                s = AttrSpec(d, self.managed_annotations, self.ignore_unmanaged, self.ignored_annotations)
                self.add_spec(s)


class AttrSpec(BaseSpec):
    def __init__(self, specdict, managed_annotations, ignore_unmanaged, ignored_annotations):
        BaseSpec.__init__(self, specdict, ["uri", "value"], "attributes", ignore_unmanaged)
        self.ignore_unmanaged = ignore_unmanaged
        self.managed_annotations = managed_annotations
        self.ignored_annotations = ignored_annotations
        self.known_annotations = self.managed_annotations + self.ignored_annotations
        self.validate_annotation()

    def validate_annotation(self):
        return self.specdict.get("uri") in self.managed_annotations


class AttrConfig:
    def __init__(self, server, catalog_id, config_file, credentials, verbose=False, schema_name=None, table_name=None):
        self.config = json.load(open(config_file))
        self.ignored_schema_patterns = []
        ip = self.config.get("ignored_schema_patterns")
        if ip is not None:
            for p in ip:
                self.ignored_schema_patterns.append(re.compile(p))
        self.known_attrs = self.config.get(u'known_attributes')
        self.managed_annotations = self.known_attrs.get(u'managed')
        self.known_annotations = self.managed_annotations
        self.all_annotations = self.known_annotations
        self.ignored_annotations = self.known_attrs.get(u'ignored')
        if self.ignored_annotations is not None:
            self.all_annotations = self.all_annotations + self.ignored_annotations
        self.ignore_unmanaged = self.known_attrs.get(u'ignore_all_unmanaged')
        self.annotation_specs = dict()
        for key in AttrSpecList.SPEC_TYPES:
            self.annotation_specs[key] = self.make_speclist(key)
        self.server = server
        self.catalog_id = catalog_id
        self.verbose = verbose
        old_catalog = ErmrestCatalog('https', self.server, self.catalog_id, credentials)
        self.saved_toplevel_config = ConfigUtil.find_toplevel_node(old_catalog.getCatalogModel(), schema_name,
                                                                   table_name)
        self.catalog = ErmrestCatalog('https', self.server, self.catalog_id, credentials)
        self.toplevel_config = ConfigUtil.find_toplevel_node(self.catalog.getCatalogModel(), schema_name, table_name)

    def make_speclist(self, name):
        d = self.config.get(unicode(name))
        if d is None:
            d = [dict()]
        return AttrSpecList(self.known_attrs, d)

    def find_best_schema_specs(self, schema_name):
        specs = dict()
        for key in self.managed_annotations:
            specs[key] = self.annotation_specs["schema_annotations"].find_best_schema_spec(schema_name, key=key)
        return specs

    def find_best_table_specs(self, schema_name, table_name):
        specs = dict()
        for key in self.managed_annotations:
            specs[key] = self.annotation_specs["table_annotations"].find_best_table_spec(schema_name, table_name,
                                                                                         key=key)
        return specs

    def find_best_fkey_specs(self, fkey):
        specs = dict()
        for key in self.managed_annotations:
            specs[key] = self.annotation_specs["foreign_key_annotations"].find_best_foreign_key_spec(fkey.table.schema.name,
                                                                                                     fkey.table.name,
                                                                                                     fkey.names,
                                                                                                     key=key)
        return specs

    def find_best_column_specs(self, schema_name, table_name, column_name):
        specs = dict()
        for key in self.managed_annotations:
            specs[key] = self.annotation_specs["column_annotations"].find_best_column_spec(schema_name, table_name,
                                                                                           column_name, key=key)
        return specs

    def node_name(self, node):
        if isinstance(node, ermrest_model.Schema):
            return "schema {s}".format(s=str(node.name))
        if isinstance(node, ermrest_model.Table):
            return "table {s}.{t}".format(s=str(node.schema.name), t=str(node.name))
        if isinstance(node, ermrest_model.Column):
            return "column {s}.{t}.{c}".format(s=str(node.table.schema.name), t=str(node.table.name), c=str(node.name))
        if isinstance(node, ermrest_model.ForeignKey):
            return "foreign key {n}".format(n=str(node.names))
        return str("unknown node type {t}".format(t=type(node)))

    def set_node_annotations(self, node, specs, saved_node):
        if specs is None:
            if not self.ignore_unmanaged:
                if self.verbose:
                    print("{n}: clearing annotations".format(n=self.node_name(node)))
                node.annotations.clear()
            return
        for k in self.managed_annotations:
            s = specs.get(k)
            if s is not None and u'value' in s:
                if self.verbose:
                    print("{n}: setting {k} to {v}".format(n=self.node_name(node), k=k, v=s[u'value']))
                node.annotations[k] = s[u'value']
            elif k in node.annotations:
                if self.verbose:
                    print("{n}: clearing {k}".format(n=self.node_name(node), k=k))
                node.annotations.pop(k)
        if not self.ignore_unmanaged:
            for k in node.annotations.keys():
                if k not in self.all_annotations:
                    raise ValueError("annotation key {k} is neither managed nor ignored".format(k=k))

    def set_table_annotations(self, table, saved_table):
        self.set_node_annotations(table, self.find_best_table_specs(table.schema.name, table.name), saved_table)
        for column in table.column_definitions:
            self.set_column_annotations(column, self.find_named_column(saved_table, column.name))
        for fkey in table.foreign_keys:
            self.set_fkey_annotations(fkey, self.find_corresponding_fkey(saved_table, fkey))

    def find_corresponding_fkey(self, table, base_fkey):
        if table is None:
            return None
        if base_fkey.names is None or len(base_fkey.names) == 0:
            return None
        names = base_fkey.names[0]
        if len(names) != 2:
            return None
        for fkey in table.foreign_keys:
            if fkey is not None and fkey.names is not None and len(fkey.names) > 0:
                for n in fkey.names:
                    if len(n) == 2 and n[0] == names[0] and n[1] == names[1]:
                        return fkey
        return None

    def find_named_column(self, table, column_name):
        if table is None:
            return None
        for column in table.column_definitions:
            if column.name == column_name:
                return column
        return None

    def find_named_schema(self, catalog, schema_name):
        if catalog is None or catalog.schemas is None:
            return None
        return catalog.schemas.get(schema_name)

    def find_named_table(self, schema, table_name):
        if schema is None:
            return None
        if schema.tables is None:
            return None
        return schema.tables.get(table_name)

    def set_fkey_annotations(self, fkey, saved_fkey):
        self.set_node_annotations(fkey, self.find_best_fkey_specs(fkey), saved_fkey)

    def set_column_annotations(self, column, saved_column):
        self.set_node_annotations(column, self.find_best_column_specs(column.table.schema.name, column.table.name, column.name),
                                  saved_column)

    def set_schema_annotations(self, schema, saved_schema):
        for pat in self.ignored_schema_patterns:
            if pat.match(schema.name) is not None:
                print("ignoring schema {s}".format(s=schema.name))
                return
        specs = self.find_best_schema_specs(schema.name)
        self.set_node_annotations(schema, specs, saved_schema)
        for table in schema.tables.values():
            self.set_table_annotations(table, self.find_named_table(saved_schema, table.name))

    def set_catalog_annotations(self):
        specs = dict()
        for key in self.managed_annotations:
            specs[key] = self.annotation_specs["catalog_annotations"].find_catalog_spec(key)
        self.set_node_annotations(self.toplevel_config, specs, self.saved_toplevel_config)

        for schema in self.toplevel_config.schemas.values():
            self.set_schema_annotations(schema, self.find_named_schema(self.saved_toplevel_config, schema.name))

    def set_attributes(self):
        if isinstance(self.toplevel_config, ermrest_model.Model):
            self.set_catalog_annotations()
        elif isinstance(self.toplevel_config, ermrest_model.Schema):
            self.set_schema_annotations(self.toplevel_config, self.saved_toplevel_config)
        elif isinstance(self.toplevel_config, ermrest_model.Table):
            self.set_table_annotations(self.toplevel_config, self.saved_toplevel_config)
        else:
            raise ValueError("toplevel config is a {t}".format(t=str(type(self.toplevel_config))))

    def apply_annotations(self):
        self.toplevel_config.apply(self.saved_toplevel_config)


def main():
    cli = ConfigBaseCLI("annotation config tool", None, version=MY_VERSION)
    args = cli.parse_cli()
    table_name = cli.get_table_arg(args)
    schema_names = cli.get_schema_arg_list(args)
    credentials = get_credential(args.host, args.credential_file)
    for schema in schema_names:
        attr_config = AttrConfig(args.host, args.catalog, args.config_file, credentials, args.verbose or args.debug,
                                 schema, table_name)
        attr_config.set_attributes()
        if not args.dryrun:
            attr_config.apply_annotations()


if __name__ == '__main__':
    sys.exit(main())



import os
import io
import re
import logging
import datetime
import codecs
import csv
import json
import requests
from typing import NamedTuple

from . import urlquote, urlsplit, urlunsplit, datapath, DEFAULT_HEADERS, DEFAULT_CHUNK_SIZE, DEFAULT_SESSION_CONFIG, \
    Megabyte, Kilobyte, get_transfer_summary
from .deriva_binding import DerivaBinding, DerivaPathError
from . import ermrest_model
from .ermrest_model import nochange

class DerivaServer (DerivaBinding):
    """Persistent handle for a Deriva server."""

    def __init__(self, scheme, server, credentials=None, caching=True, session_config=None):
        """Create a Deriva server binding.

           Arguments:
             scheme: 'http' or 'https'
             server: server FQDN string
             credentials: credential secrets, e.g. cookie
             caching: whether to retain a GET response cache
        """
        super(DerivaServer, self).__init__(scheme, server, credentials, caching, session_config)
        self.scheme = scheme
        self.server = server
        self.credentials = credentials
        self.caching = caching
        self.session_config = session_config

    def connect_ermrest(self, catalog_id, snaptime=None):
        """Connect to an ERMrest catalog and return the catalog binding.

        :param catalog_id: The id (or alias) of the existing catalog
        :param snaptime: The id for a desired catalog snapshot (default None)

        The catalog_id is normally a bare id (str), and the optional
        snaptime is a bare snapshot id (str). If the snaptime is None,
        the catalog_id may be a concatenated <id>@<snaptime> string,
        and it will be split to determine the snaptime.

        If no snaptime is passed separately or compounded with
        catalog_id, an ErmrestCatalog binding will be
        returned. Conversely, if a snaptime is determined, an
        ErmrestSnapshot (immutable) binding will be returned.

        """
        return ErmrestCatalog.connect(self, catalog_id, snaptime)

    def create_ermrest_catalog(self, id=None, owner=None):
        """Create an ERMrest catalog.

        :param id: The (str) id desired by the client (default None)
        :param owner: The initial (list of str) ACL desired by the client (default None)

        The new catalog id will be returned in the response, and used
        in future catalog access. The use of the id parameter
        may yield errors if the supplied value is not available for
        use by the client. The value None will result in a
        server-assigned catalog id.

        The initial "owner" ACL on the new catalog will be the
        client-supplied owner if provided. The use of owner parameter
        may yield errors if the supplied ACL does not match the
        client, i.e. the client cannot lock themselves out of the
        catalog. The value None will result in a server-assigned ACL
        with the requesting client's identity.

        Certain failure modes (or message loss) may leave the
        id reserved in the system. In this case, the effective
        owner ACL influences which client(s) are allowed to retry
        creation with the same id.

        """
        return ErmrestCatalog.create(self, id, owner)

    def connect_ermrest_alias(self, id):
        """Connect to an ERMrest alias and return the alias binding.

        :param id: The id of the existing alias

        """
        return ErmrestAlias.connect(self, id)

    def create_ermrest_alias(self, id=None, owner=None, alias_target=None):
        """Create an ERMrest catalog alias.

        :param id: The (str) id desired by the client (default None)
        :param owner: The initial (list of str) ACL desired by the client (default None)
        :param alias_target: The initial target catalog id binding desired by the client (default None)

        The new alias id will be returned in the response, and used
        in future alias access. The use of the id parameter
        may yield errors if the supplied value is not available for
        use by the client. The value None will result in a
        server-assigned alias id.

        The initial "owner" ACL on the new alias will be the
        client-supplied owner. The use of owner parameter may yield
        errors if the supplied ACL does not match the client, i.e. the
        client cannot lock themselves out of the alias. The value
        None will result in a server-assigned ACL with the requesting
        client's identity.

        The alias is bound to the client-supplied alias_target, if
        supplied. The use of alias_target may yield errors if the
        supplied value is not a valid target catalog id. The value
        None will reserve the alias in an unbound state.

        Certain failure modes (or message loss) may leave the id
        reserved in the system. In this case, the effective owner_acl
        influences which client(s) are allowed to retry creation with
        the same id.

        """
        return ErmrestAlias.create(self, id, owner, alias_target)

class ErmrestCatalogMutationError(Exception):
    pass


_clone_state_url = "tag:isrd.isi.edu,2018:clone-status"

DEFAULT_PAGE_SIZE = 100000

class ResolveRidResult (NamedTuple):
    datapath: datapath.DataPath
    table: ermrest_model.Table
    rid: str


class ErmrestCatalog(DerivaBinding):
    """Persistent handle for an ERMrest catalog.

       Provides basic REST client for HTTP methods on arbitrary
       paths. Caller has to understand ERMrest APIs and compose
       appropriate paths, headers, and/or content.

       Additional utility methods provided for accessing catalog metadata.
    """
    table_schemas = dict()

    @property
    def deriva_server(self):
        """Return DerivaServer binding for the same server this catalog belongs to."""
        return DerivaServer(
            self._scheme,
            self._server,
            self._credentials,
            self._caching,
            self._session_config,
        )

    @classmethod
    def connect(cls, deriva_server, catalog_id, snaptime=None):
        """Connect to an ERMrest catalog and return the catalog binding.

        :param deriva_server: The DerivaServer binding which hosts ermrest
        :param catalog_id: The id (or alias) of the existing catalog
        :param snaptime: The id for a desired catalog snapshot (default None)

        The catalog_id is normally a bare id (str), and the optional
        snaptime is a bare snapshot id (str). If the snaptime is None,
        the catalog_id may be a concatenated <id>@<snaptime> string,
        and it will be split to determine the snaptime.

        If no snaptime is passed separately or compounded with
        catalog_id, an ErmrestCatalog binding will be
        returned. Conversely, if a snaptime is determined, an
        ErmrestSnapshot (immutable) binding will be returned.

        """
        if not snaptime:
            splits = str(catalog_id).split('@')
            if len(splits) > 2:
                raise Exception('Malformed catalog identifier: multiple "@" characters found.')
            catalog_id = splits[0]
            snaptime = splits[1] if len(splits) == 2 else None

        if snaptime:
            return ErmrestSnapshot(
                deriva_server.scheme,
                deriva_server.server,
                catalog_id,
                snaptime,
                deriva_server.credentials,
                deriva_server.caching,
                deriva_server.session_config
            )

        return cls(
            deriva_server.scheme,
            deriva_server.server,
            catalog_id,
            deriva_server.credentials,
            deriva_server.caching,
            deriva_server.session_config
        )

    @classmethod
    def _digest_catalog_args(cls, id, owner):
        rep = dict()

        if isinstance(id, str):
            rep['id'] = id
        elif isinstance(id, (type(nochange), type(None))):
            pass
        else:
            raise TypeError('id must be of type str or None or nochange, not %s' % type(id))

        if isinstance(owner, list):
            for e in owner:
                if not isinstance(e, str):
                    raise TypeError('owner members must be of type str, not %s' % type(e))
            rep['owner'] = owner
        elif isinstance(owner, (type(nochange), type(None))):
            pass
        else:
            raise TypeError('owner must be of type list or None or nochange, not %s' % type(owner))

        return rep

    @classmethod
    def create(cls, deriva_server, id=None, owner=None):
        """Create an ERMrest catalog and return the ERMrest catalog binding.

        :param deriva_server: The DerivaServer binding which hosts ermrest.
        :param id: The (str) id desired by the client (default None)
        :param owner: The initial (list of str) ACL desired by the client (default None)

        The new catalog id will be returned in the response, and used
        in future catalog access. The use of the id parameter
        may yield errors if the supplied value is not available for
        use by the client. The value None will result in a
        server-assigned catalog id.

        The initial "owner" ACL on the new catalog will be the
        client-supplied owner ACL. The use of owner parameter
        may yield errors if the supplied ACL does not match the
        client, i.e. the client cannot lock themselves out of the
        catalog. The value None will result in a server-assigned ACL
        with the requesting client's identity.

        Certain failure modes (or message loss) may leave the id
        reserved in the system. In this case, the effective owner ACL
        influences which client(s) are allowed to retry creation with
        the same id.

        """
        path = '/ermrest/catalog'
        r = deriva_server.post(path, json=cls._digest_catalog_args(id, owner))
        r.raise_for_status()
        return cls.connect(deriva_server, r.json()['id'])

    def __init__(self, scheme, server, catalog_id, credentials=None, caching=True, session_config=None):
        """Create ERMrest catalog binding.

           Arguments:
             scheme: 'http' or 'https'
             server: server FQDN string
             catalog_id: e.g. '1'
             credentials: credential secrets, e.g. cookie
             caching: whether to retain a GET response cache

           Deriva Client Context: You MAY mutate self.dcctx to
           customize the context for this service endpoint prior to
           invoking web requests.  E.g.:

             self.dcctx['cid'] = 'my application name'

           You MAY also supply custom per-request context by passing a
           headers dict to web request methods, e.g.

             self.get(..., headers={'deriva-client-context': {'action': 'myapp/function1'}})

           This custom header will be merged as override values with
           the default context in self.dcctx in order to form the
           complete context for the request.
        """
        super(ErmrestCatalog, self).__init__(scheme, server, credentials, caching, session_config)
        if isinstance(catalog_id, int):
            catalog_id = str(catalog_id)
        self._server_uri = "%s/ermrest/catalog/%s" % (
            self._server_uri,
            urlquote(catalog_id),
        )
        self._scheme, self._server, self._catalog_id, self._credentials, self._caching, self._session_config = \
            scheme, server, catalog_id, credentials, caching, session_config

    @property
    def catalog_id(self):
        return self._catalog_id

    @property
    def alias_target(self):
        r = self.get('/')
        r.raise_for_status()
        rep = r.json()
        return rep.get('alias_target')

    def exists(self):
        """Simple boolean test for catalog existence.

        :return: True if exists, False if not (404), otherwise raises exception
        """
        try:
            self.get('/')
            return True
        except requests.HTTPError as e:
            if e.response.status_code == 404:
                return False
            else:
                raise

    def latest_snapshot(self):
        """Gets a handle to this catalog's latest snapshot.
        """
        r = self.get('/')
        r.raise_for_status()
        return ErmrestSnapshot(self._scheme, self._server, self._catalog_id, r.json()['snaptime'],
                               self._credentials, self._caching, self._session_config)

    def getCatalogModel(self):
        return ermrest_model.Model.fromcatalog(self)

    def getCatalogSchema(self):
        path = '/schema'
        r = self.get(path)
        r.raise_for_status()
        return r.json()

    def getPathBuilder(self):
        """Returns the 'path builder' interface for this catalog."""
        return datapath.from_catalog(self)

    def getTableSchema(self, fq_table_name):
        # first try to get from cache(s)
        s, t = self.splitQualifiedCatalogName(fq_table_name)
        cat = self.getCatalogSchema()
        schema = cat['schemas'][s]['tables'][t] if cat else None
        if schema:
            return schema
        schema = self.table_schemas.get(fq_table_name)
        if schema:
            return schema

        path = '/schema/%s/table/%s' % (s, t)
        r = self.get(path)
        resp = r.json()
        self.table_schemas[fq_table_name] = resp
        r.raise_for_status()

        return resp

    def getTableColumns(self, fq_table_name):
        columns = set()
        schema = self.getTableSchema(fq_table_name)
        for column in schema['column_definitions']:
            columns.add(column['name'])

        return columns

    def validateRowColumns(self, row, fq_tableName):
        columns = self.getTableColumns(fq_tableName)
        return set(row.keys()) - columns

    def getDefaultColumns(self, row, table, exclude=None, quote_url=True):
        columns = self.getTableColumns(table)
        if isinstance(exclude, list):
            for col in exclude:
                columns.remove(col)

        defaults = []
        supplied_columns = row.keys()
        for col in columns:
            if col not in supplied_columns:
                defaults.append(urlquote(col, safe='') if quote_url else col)

        return defaults

    @staticmethod
    def splitQualifiedCatalogName(name):
        entity = name.split(':')
        if len(entity) != 2:
            logging.debug("Unable to tokenize %s into a fully qualified <schema:table> name." % name)
            return None
        return entity[0], entity[1]

    def resolve_rid(self, rid: str, model: ermrest_model.Model=None, builder: datapath._CatalogWrapper=None) -> ResolveRidResult:
        """Resolve a RID value to return a ResolveRidResult (a named tuple).

        :param rid: The RID (str) to resolve
        :param model: A result from self.getCatalogModel() to reuse
        :param builder: A result from self.getPathBuilder() to reuse

        Raises KeyError if RID is not found in the catalog.

        The elements of the ResolveRidResult namedtuple provide more
        information about the entity identified by the supplied RID in
        this catalog:

        - datapath: datapath instance for querying the resolved entity
        - table: ermrest_model.Table instance containing the entity
        - rid: normalized version of the input RID value

        Example to simply retrieve entity content:

           path, _, _ = catalog.resolve_rid('1-0000')
           data = path.entities().fetch()[0]

        """
        if model is None:
            model = self.getCatalogModel()
        if builder is None:
            builder = self.getPathBuilder()
        try:
            r = self.get('/entity_rid/%s' % urlquote(rid))
            info = r.json()
            sname = info['schema_name']
            tname = info['table_name']
            rid = info['RID']

            ptable = builder.schemas[sname].tables[tname]

            return ResolveRidResult(
                ptable.path.filter(ptable.RID == rid),
                model.schemas[sname].tables[tname],
                rid
            )
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                raise KeyError(rid)
            raise

    def getAsFile(self,
                  path,
                  destfilename,
                  headers=DEFAULT_HEADERS,
                  callback=None,
                  delete_if_empty=False,
                  paged=False,
                  page_size=DEFAULT_PAGE_SIZE,
                  page_sort_columns=frozenset(["RID"])):
        """
           Retrieve catalog data streamed to destination file.
           Caller is responsible to clean up file even on error, when the file may or may not exist.
           If "delete_if_empty" is True, the file will be inspected for "empty" content. In the case of
           json/json-stream content, the presence of a single empty JSON object will be tested for. In the case of
           CSV content, the file will be parsed with CSV reader to determine that only a single header line and no row
           data is present.
        """
        self.check_path(path)

        # Only entity API supported with paged mode at this time, otherwise fallback. We fallback rather than raise an
        # exception in the case that the caller might be trying to perform an opportunistic paged request without
        # knowing a priori if paged support for the given query is available.
        page_size = page_size if page_size > 0 else DEFAULT_PAGE_SIZE
        if not (path.startswith("/entity") or path.startswith("/attribute")) and paged:
            logging.warning("Paged data retrieval only supported for entity or attribute API queries.")
            paged = False

        # Only "application/x-json-stream" or "text/csv" supported with paged mode at this time, otherwise fallback.
        accept = headers.get("accept")
        if accept not in ("application/x-json-stream", "text/csv"):
            logging.debug("Paged data retrieval not supported for content type: %s" % accept)
            paged = False

        headers = headers.copy()

        destfile = open(destfilename, 'w+b')

        try:
            total = 0
            start = datetime.datetime.now()

            if not paged:
                with self._session.get(self._server_uri + path, headers=headers, stream=True) as r:
                    self._response_raise_for_status(r)
                    content_type = r.headers.get("Content-Type")
                    logging.debug("Transferring file %s to %s" % (self._server_uri + path, destfilename))
                    for buf in r.iter_content(chunk_size=DEFAULT_CHUNK_SIZE):
                        destfile.write(buf)
                        total += len(buf)
                        if callback:
                            if not callback(progress="Downloading: %.2f MB transferred" %
                                                     (float(total) / float(Megabyte))):
                                destfile.close()
                                return
                destfile.flush()
            else:
                first_page = True
                first_line = None
                last_record = None
                usr = urlsplit(self._server_uri + path)
                path = str(usr.path.split('@sort')[0])
                while True:
                    sort = "@sort(%s)%s" % (",".join(page_sort_columns or ["RID"]),
                                            ("@after(%s)" % ",".join(last_record)) if last_record is not None else "")
                    limit = "limit=%s" % int(page_size) if page_size > 0 else "none"
                    query = re.sub(r"([^.]*)(limit=.*?)($|[&;])([^.]*)$", r"\1%s\3\4" % limit, usr.query, flags=re.I)
                    url = urlunsplit((usr.scheme, usr.netloc, path + sort, query if query else limit, usr.fragment))

                    # 1. Try to get a page worth of data, back-off page size if query run time errors are encountered
                    with self._session.get(url, headers=headers) as r:
                        if r.status_code == 400 and "Query run time limit exceeded" in r.text:
                            if page_size == 1:
                                self._response_raise_for_status(r)
                            r.close()
                            page_size //= 2
                            page_size = 1 if page_size < 1 else page_size
                            logging.warning("Query runtime exceeded while attempting to transfer rows from %s to file "
                                            "[%s]. The page size is being reduced to %s and the query will be retried."
                                            % (url, destfilename, page_size))
                            if callback:
                                if not callback(progress="Retrying query: %s" % url):
                                    destfile.close()
                                    return
                            continue
                        else:
                            self._response_raise_for_status(r)

                        # 2. Write the page to disk and check the last record processed in order to get the next page
                        last_line = {}
                        content_type = r.headers.get("Content-Type")
                        logging.debug("Transferring data from [%s] to %s" % (url, destfilename))
                        # CSV processing iterates over lines in the response, skipping the header line(s) in all but
                        # the first page, and captures the last line of each page to determine the last record processed
                        if content_type == "text/csv":
                            skip = 1
                            line_num = 0
                            if first_page:
                                lines = r.iter_lines(decode_unicode=True)
                                reader = csv.reader(lines)
                                first_line = next(reader)
                                skip = reader.line_num
                            for line in r.iter_lines():
                                if not first_page:
                                    line_num += 1
                                    if line_num <= skip:
                                        continue
                                tline = line + b"\n"
                                destfile.write(tline)
                                total += len(tline)
                                last_line = tline
                            if last_line and last_line != first_line:
                                reader = csv.DictReader([last_line.decode('utf-8')], first_line)
                                last_line = next(reader)
                            first_page = False
                        # JSON-Stream processing writes the entire buffer to the destination file. The last line is
                        # captured by reverse seeking in the buffer from right before the last b'\n' newline to the next
                        # newline or buf[0], then calling readline from the current position
                        elif content_type == "application/x-json-stream":
                            buf = r.content
                            if not buf:
                                break
                            destfile.write(buf)
                            total += len(buf)
                            b = io.BytesIO(buf)
                            b.seek(-2, os.SEEK_END)
                            while b.read(1) != b'\n':
                                b.seek(-2, os.SEEK_CUR)
                                if b.tell() == os.SEEK_SET:
                                    break
                            last_line = json.loads(b.readline().decode('utf-8'))

                        # 3. Save the last record key and flush the destination file buffers to disk.
                        if not last_line:
                            break
                        destfile.flush()
                        last_record = [urlquote(str(last_line.get(key))) for key in page_sort_columns]
                        if callback:
                            if not callback(progress="Downloading: %.2f MB transferred" %
                                                     (float(total) / float(Megabyte))):
                                destfile.close()
                                return

            elapsed = datetime.datetime.now() - start
            summary = get_transfer_summary(total, elapsed)

            # perform automatic file deletion on detected "empty" content, if requested
            delete_file = True if total == 0 else False
            if delete_if_empty and total > 0:
                destfile.seek(0)
                if content_type == "application/json" or content_type == "application/x-json-stream":
                    buf = destfile.read(16)
                    if buf == b"[]\n" or buf == b"{}\n":
                        delete_file = True
                elif content_type == "text/csv":
                    reader = csv.reader(codecs.iterdecode(destfile, 'utf-8'))
                    rowcount = 0
                    for row in reader:
                        rowcount += 1
                        if rowcount > 1:
                            break
                    if rowcount <= 1:
                        delete_file = True

            # automatically delete zero-length files or detected "empty" content
            if delete_file:
                destfile.close()
                os.remove(destfilename)
                destfile = None

            log_msg = "File [%s] transfer successful. %s %s" % \
                      (destfilename, summary,
                       "File was automatically deleted due to empty content." if delete_file else "")
            logging.info(log_msg)
            if callback:
                callback(summary=log_msg, file_path=destfilename)

        finally:
            if destfile:
                destfile.close()

    def delete(self, path, headers=DEFAULT_HEADERS, guard_response=None):
        """Perform DELETE request, returning response object.

           Arguments:
             path: the path within this bound catalog
             headers: headers to set in request
             guard_response: expected current resource state
               as previously seen response object.

           Uses guard_response to build appropriate 'if-match' header
           to assure change is only applied to expected state.

           Raises ConcurrentUpdate for 412 status.

        """
        if path == "/":
            raise DerivaPathError('See self.delete_ermrest_catalog() if you really want to destroy this catalog.')
        return DerivaBinding.delete(self, path, headers=headers, guard_response=guard_response)

    def delete_ermrest_catalog(self, really=False):
        """Perform DELETE request, destroying catalog on server.

           Arguments:
             really: delete when True, abort when False (default)

        """
        if really is True:
            return DerivaBinding.delete(self, '/')
        else:
            raise ValueError('Catalog deletion refused when really is %s.' % really)

    def clone_catalog(self,
                      dst_catalog=None,
                      copy_data=True,
                      copy_annotations=True,
                      copy_policy=True,
                      truncate_after=True,
                      exclude_schemas=None):
        """Clone this catalog's content into dest_catalog, creating a new catalog if needed.

        :param dst_catalog: Destination catalog or None to request creation of new destination (default).
        :param copy_data: Copy table contents when True (default).
        :param copy_annotations: Copy annotations when True (default).
        :param copy_policy: Copy access-control policies when True (default).
        :param truncate_after: Truncate destination history after cloning when True (default).
        :param exclude_schemas: A list of schema names to exclude from the cloning process.

        When dest_catalog is provided, attempt an idempotent clone,
        assuming content MAY be partially cloned already using the
        same parameters. This routine uses a table-level annotation
        "tag:isrd.isi.edu,2018:clone-state" to save progress markers
        which help it restart efficiently if interrupted.

        Cloning preserves source row RID values for application tables
        so that any RID-based foreign keys are still valid. It is not
        generally advisable to try to merge more than one source into
        the same clone, nor to clone on top of rows generated locally
        in the destination, since this could cause duplicate RID
        conflicts.

        Cloning does not preserve all RID values for special ERMrest
        tables in the public schema (e.g. ERMrest_Client,
        ERMrest_Group) but normal applications should only consider
        the ID key of these tables.

        Truncation after cloning avoids retaining incremental
        snapshots which contain partial clones.

        """
        src_model = self.getCatalogModel()
        session_config = self._session_config.copy() if self._session_config else DEFAULT_SESSION_CONFIG.copy()
        session_config["allow_retry_on_all_methods"] = True

        if dst_catalog is None:
            # TODO: refactor with DerivaServer someday
            server = DerivaBinding(self._scheme, self._server, self._credentials, self._caching, session_config)
            dst_id = server.post("/ermrest/catalog").json()["id"]
            dst_catalog = ErmrestCatalog(self._scheme, self._server, dst_id, self._credentials, self._caching, session_config)

        # set top-level config right away and find fatal usage errors...
        if copy_policy:
            if not src_model.acls:
                raise ValueError("Use of copy_policy=True not possible when caller does not own source catalog.")
            dst_catalog.put('/acl', json=src_model.acls)

        if copy_annotations:
            dst_catalog.put('/annotation', json=src_model.annotations)

        # build up the model content we will copy to destination
        dst_model = dst_catalog.getCatalogModel()

        new_model = []
        new_columns = [] # ERMrest does not currently allow bulk column creation
        new_keys = [] # ERMrest does not currently allow bulk key creation
        clone_states = {}
        fkeys_deferred = {}
        exclude_schemas = [] if exclude_schemas is None else exclude_schemas

        def prune_parts(d, *extra_victims):
            victims = set(extra_victims)
            # we will apply config as a second pass after extending dest model
            # but loading bulk first may speed that up
            if not copy_annotations:
                victims |= {'annotations',}
            if not copy_policy:
                victims |= {'acls', 'acl_bindings'}
            for k in victims:
                d.pop(k, None)
            return d

        def copy_sdef(s):
            """Copy schema definition structure with conditional parts for cloning."""
            d = prune_parts(s.prejson(), 'tables')
            return d

        def copy_tdef_core(t):
            """Copy table definition structure with conditional parts excluding fkeys."""
            d = prune_parts(t.prejson(), 'foreign_keys')
            d['column_definitions'] = [ prune_parts(c) for c in d['column_definitions'] ]
            d['keys'] = [ prune_parts(c) for c in d.get('keys', []) ]
            d.setdefault('annotations', {})[_clone_state_url] = 1 if copy_data else None
            return d

        def copy_tdef_fkeys(t):
            """Copy table fkeys structure."""
            def check(fkdef):
                for fkc in fkdef['referenced_columns']:
                    if fkc['schema_name'] == 'public' \
                       and fkc['table_name'] in {'ERMrest_Client', 'ERMrest_Group', 'ERMrest_RID_Lease'} \
                       and fkc['column_name'] == 'RID':
                        raise ValueError("Cannot clone catalog with foreign key reference to %(schema_name)s:%(table_name)s:%(column_name)s" % fkc)
                return fkdef
            return [ prune_parts(check(d)) for d in t.prejson().get('foreign_keys', []) ]

        def copy_cdef(c):
            """Copy column definition with conditional parts."""
            return (sname, tname, prune_parts(c.prejson()))

        def check_column_compatibility(src, dst):
            """Check compatibility of source and destination column definitions."""
            def error(fieldname, sv, dv):
                return ValueError("Source/dest column %s mismatch %s != %s for %s:%s:%s" % (
                    fieldname,
                    sv, dv,
                    src.sname, src.tname, src.name
                ))
            if src.type.typename != dst.type.typename:
                raise error("type", src.type.typename, dst.type.typename)
            if src.nullok != dst.nullok:
                raise error("nullok", src.nullok, dst.nullok)
            if src.default != dst.default:
                raise error("default", src.default, dst.default)

        def copy_kdef(k):
            return (sname, tname, prune_parts(k.prejson()))

        for sname, schema in src_model.schemas.items():
            if sname in exclude_schemas:
                continue
            if sname not in dst_model.schemas:
                new_model.append(copy_sdef(schema))

            for tname, table in schema.tables.items():
                if table.kind != 'table':
                    logging.warning('Skipping cloning of %s %s:%s' % (table.kind, sname, tname))
                    continue

                if 'RID' not in table.column_definitions.elements:
                    raise ValueError("Source table %s.%s lacks system-columns and cannot be cloned." % (sname, tname))

                if sname not in dst_model.schemas or tname not in dst_model.schemas[sname].tables:
                    new_model.append(copy_tdef_core(table))
                    clone_states[(sname, tname)] = 1 if copy_data else None
                    fkeys_deferred[(sname, tname)] = copy_tdef_fkeys(table)
                else:
                    if dst_model.schemas[sname].tables[tname].foreign_keys:
                        # assume that presence of any destination foreign keys means we already loaded deferred_fkeys
                        copy_data = False
                    else:
                        fkeys_deferred[(sname, tname)] = copy_tdef_fkeys(table)

                    src_columns = { c.name: c for c in table.column_definitions }
                    dst_columns = { c.name: c for c in dst_model.schemas[sname].tables[tname].column_definitions }

                    for cname in src_columns:
                        if cname not in dst_columns:
                            new_columns.append(copy_cdef(src_columns[cname]))
                        else:
                            check_column_compatibility(src_columns[cname], dst_columns[cname])

                    for cname in dst_columns:
                        if cname not in src_columns:
                            raise ValueError("Destination column %s.%s.%s does not exist in source catalog." % (sname, tname, cname))

                    src_keys = { tuple(sorted(c.name for c in key.unique_columns)): key for key in table.keys }
                    dst_keys = { tuple(sorted(c.name for c in key.unique_columns)): key for key in dst_model.schemas[sname].tables[tname].keys }

                    for utuple in src_keys:
                        if utuple not in dst_keys:
                            new_keys.append(copy_kdef(src_keys[utuple]))

                    for utuple in dst_keys:
                        if utuple not in src_keys:
                            raise ValueError("Destination key %s.%s(%s) does not exist in source catalog." % (sname, tname, ', '.join(utuple)))

                    clone_states[(sname, tname)] = dst_model.schemas[sname].tables[tname].annotations.get(_clone_state_url)

        clone_states[('public', 'ERMrest_RID_Lease')] = None # never try to sync leases

        # apply the stage 1 model to the destination in bulk
        if new_model:
            dst_catalog.post("/schema", json=new_model).raise_for_status()

        for sname, tname, cdef in new_columns:
            dst_catalog.post("/schema/%s/table/%s/column" % (urlquote(sname), urlquote(tname)), json=cdef).raise_for_status()

        for sname, tname, kdef in new_keys:
            dst_catalog.post("/schema/%s/table/%s/key" % (urlquote(sname), urlquote(tname)), json=kdef).raise_for_status()

        # copy data in stage 2
        if copy_data:
            page_size = 10000
            for sname, tname in clone_states.keys():
                tname_uri = "%s:%s" % (urlquote(sname), urlquote(tname))
                if clone_states[(sname, tname)] == 1:
                    # determine current position in (partial?) copy
                    r = dst_catalog.get("/entity/%s@sort(RID::desc::)?limit=1" % tname_uri).json()
                    if r:
                        last = r[0]['RID']
                    else:
                        last = None

                    while True:
                        page = self.get(
                            "/entity/%s@sort(RID)%s?limit=%d" % (
                                tname_uri,
                                ("@after(%s)" % urlquote(last)) if last is not None else "",
                                page_size
                            )
                        ).json()
                        if page:
                            dst_catalog.post("/entity/%s?nondefaults=RID,RCT,RCB" % tname_uri, json=page)
                            last = page[-1]['RID']
                        else:
                            break

                    # record our progress on catalog in case we fail part way through
                    dst_catalog.put(
                        "/schema/%s/table/%s/annotation/%s" % (
                            urlquote(sname),
                            urlquote(tname),
                            urlquote(_clone_state_url),
                        ),
                        json=2
                    )
                elif clone_states[(sname, tname)] is None and (sname, tname) in {
                        ('public', 'ERMrest_Client'),
                        ('public', 'ERMrest_Group'),
                }:
                    # special sync behavior for magic ermrest tables
                    # HACK: these are assumed small enough to join via local merge of arrays
                    page = self.get("/entity/%s?limit=none" % tname_uri).json()
                    dst_catalog.post("/entity/%s?onconflict=skip" % tname_uri, json=page)

                    # record our progress on catalog in case we fail part way through
                    dst_catalog.put(
                        "/schema/%s/table/%s/annotation/%s" % (
                            urlquote(sname),
                            urlquote(tname),
                            urlquote(_clone_state_url),
                        ),
                        json=2
                    )

        # apply stage 2 model in bulk only... we won't get here unless preceding succeeded
        new_fkeys = []
        for fkeys in fkeys_deferred.values():
            new_fkeys.extend(fkeys)

        if new_fkeys:
            dst_catalog.post("/schema", json=new_fkeys)

        # copy over configuration in stage 3
        # we need to do this after deferred_fkeys to handle acl_bindings projections with joins
        dst_model = dst_catalog.getCatalogModel()

        for sname, src_schema in src_model.schemas.items():
            if sname in exclude_schemas:
                continue
            dst_schema = dst_model.schemas[sname]

            if copy_annotations:
                dst_schema.annotations.clear()
                dst_schema.annotations.update(src_schema.annotations)

            if copy_policy:
                dst_schema.acls.clear()
                dst_schema.acls.update(src_schema.acls)

            for tname, src_table in src_schema.tables.items():
                dst_table = dst_schema.tables[tname]

                if copy_annotations:
                    merged = dict(src_table.annotations)
                    if _clone_state_url in dst_table.annotations:
                        merged[_clone_state_url] = dst_table.annotations[_clone_state_url]
                    dst_table.annotations.clear()
                    dst_table.annotations.update(merged)

                if copy_policy:
                    dst_table.acls.clear()
                    dst_table.acls.update(src_table.acls)
                    dst_table.acl_bindings.clear()
                    dst_table.acl_bindings.update(src_table.acl_bindings)

                for cname, src_col in src_table.columns.elements.items():
                    dst_col = dst_table.columns[cname]

                    if copy_annotations:
                        dst_col.annotations.clear()
                        dst_col.annotations.update(src_col.annotations)

                    if copy_policy:
                        dst_col.acls.clear()
                        dst_col.acls.update(src_col.acls)
                        dst_col.acl_bindings.clear()
                        dst_col.acl_bindings.update(src_col.acl_bindings)

                for src_key in src_table.keys:
                    dst_key = dst_table.key_by_columns([ col.name for col in src_key.unique_columns ])

                    if copy_annotations:
                        dst_key.annotations.clear()
                        dst_key.annotations.update(src_key.annotations)

                def xlate_column_map(fkey):
                    dst_from_table = dst_table
                    dst_to_schema = dst_model.schemas[fkey.pk_table.schema.name]
                    dst_to_table = dst_to_schema.tables[fkey.pk_table.name]
                    return {
                        dst_from_table._own_column(from_col.name): dst_to_table._own_column(to_col.name)
                        for from_col, to_col in fkey.column_map.items()
                    }

                for src_fkey in src_table.foreign_keys:
                    dst_fkey = dst_table.fkey_by_column_map(xlate_column_map(src_fkey))

                    if copy_annotations:
                        dst_fkey.annotations.clear()
                        dst_fkey.annotations.update(src_fkey.annotations)

                    if copy_policy:
                        dst_fkey.acls.clear()
                        dst_fkey.acls.update(src_fkey.acls)
                        dst_fkey.acl_bindings.clear()
                        dst_fkey.acl_bindings.update(src_fkey.acl_bindings)

        # send all the config changes to the server
        dst_model.apply()

        # truncate cloning history
        if truncate_after:
            snaptime = dst_catalog.get("/").json()["snaptime"]
            dst_catalog.delete("/history/,%s" % urlquote(snaptime))

        return dst_catalog

class ErmrestSnapshot(ErmrestCatalog):
    """Persistent handle for an ERMrest catalog snapshot.

    Inherits from ErmrestCatalog and provides the same interfaces,
    except that the interfaces are now bound to a fixed snapshot
    of the catalog.
    """
    def __init__(self, scheme, server, catalog_id, snaptime, credentials=None, caching=True, session_config=None):
        """Create ERMrest catalog snapshot binding.

           Arguments:
             scheme: 'http' or 'https'
             server: server FQDN string
             catalog_id: e.g., '1'
             snaptime: e.g., '2PM-DGYP-56Z4'
             credentials: credential secrets, e.g. cookie
             caching: whether to retain a GET response cache
        """
        super(ErmrestSnapshot, self).__init__(scheme, server, catalog_id, credentials, caching, session_config)
        self._server_uri = "%s@%s" % (
            self._server_uri,
            snaptime
        )
        self._snaptime = snaptime

    @property
    def snaptime(self):
        """The snaptime for this catalog snapshot instance."""
        return self._snaptime

    def _pre_mutate(self, path, headers, guard_response=None):
        """Override and disable mutation operations.

        When called by the super-class, this method raises an exception.
        """
        raise ErmrestCatalogMutationError('Catalog snapshot is immutable')

class ErmrestAlias(DerivaBinding):
    """Persistent handle for an ERMrest alias.

       Provides basic REST client for HTTP methods on arbitrary
       paths. Caller has to understand ERMrest APIs and compose
       appropriate paths, headers, and/or content.

       Additional utility methods provided for accessing alias metadata.
    """
    @classmethod
    def connect(cls, deriva_server, alias_id):
        """Connect to an ERMrest alias and return the alias binding.

        :param deriva_server: The DerivaServer binding which hosts ermrest
        :param alias_id: The id of the existing alias

        The alias_id is a bare id (str).

        """
        return cls(
            deriva_server.scheme,
            deriva_server.server,
            alias_id,
            deriva_server.credentials,
            deriva_server.caching,
            deriva_server.session_config
        )

    @classmethod
    def _digest_alias_args(cls, id, owner, alias_target):
        rep = ErmrestCatalog._digest_catalog_args(id, owner)

        if isinstance(alias_target, (str, type(None))):
            rep['alias_target'] = alias_target
        elif isinstance(alias_target, type(nochange)):
            pass
        else:
            raise TypeError('alias_target must be of type str or None or nochange, not %s' % type(alias_target))

        return rep

    @classmethod
    def create(cls, deriva_server, id=None, owner=None, alias_target=None):
        """Create an ERMrest catalog alias.

        :param deriva_server: The DerivaServer binding which hosts ermrest
        :param id: The (str) id desired by the client (default None)
        :param owner: The initial (list of str) ACL desired by the client (default None)
        :param alias_target: The initial target catalog id desired by the client (default None)

        The new alias id will be returned in the response, and used
        in future alias access. The use of the id parameter
        may yield errors if the supplied value is not available for
        use by the client. The value None will result in a
        server-assigned alias id.

        The initial "owner" ACL on the new alias will be the
        client-supplied owner parameter. The use of owner may yield
        errors if the supplied ACL does not match the client, i.e. the
        client cannot lock themselves out of the alias. The value None
        will result in a server-assigned ACL with the requesting
        client's identity.

        The alias is bound to the client-supplied alias_target, if
        supplied. The use of alias_target may yield errors if the
        supplied value is not a valid target catalog id. The value
        None will reserve the alias in an unbound state.

        Certain failure modes (or message loss) may leave the id
        reserved in the system. In this case, the effective owner ACL
        influences which client(s) are allowed to retry creation with
        the same id.

        """
        path = '/ermrest/alias'
        r = deriva_server.post(path, json=cls._digest_alias_args(id, owner, alias_target))
        r.raise_for_status()
        return cls.connect(deriva_server, r.json()['id'])

    def __init__(self, scheme, server, alias_id, credentials=None, caching=True, session_config=None):
        """Create ERMrest alias binding.

        :param scheme: 'http' or 'https'
        :param server: server FQDN string
        :param alias_id: e.g. '1'
        :param credentials: credential secrets, e.g. cookie
        :param caching: whether to retain a GET response cache

        """
        super(ErmrestAlias, self).__init__(scheme, server, credentials, caching, session_config)
        self._server_uri = "%s/ermrest/alias/%s" % (
            self._server_uri,
            alias_id
        )
        self._scheme, self._server, self._alias_id, self._credentials, self._caching, self._session_config = \
            scheme, server, alias_id, credentials, caching, session_config

    @property
    def alias_id(self):
        return self._alias_id

    def check_path(self, path):
        if path != '':
            raise ValueError('ErmrestAlias requires "" relative path')

    def retrieve(self):
        """Retrieve current alias binding state as a dict.

        The returned dictionary is suitable for local revision and
        being passed back into self.update:

           state = self.retrieve()
           state.update({ "owner": ..., "alias_target": ...)
           self.update(**state)

        """
        return self.get('').json()

    def update(self, owner=nochange, alias_target=nochange, id=None):
        """Update alias binding state in server, returning the response message dict.

        :param owner: Revised owner ACL for binding or nochange (default None)
        :param alias_target: Revised target for binding or nochange (default None)
        :param id: Current self.alias_id or None (default None)

        The optional id parameter must be None or self.alias_id and
        does not affect state changes to the server. It is only
        specified in order to allow an idiom like:

           state = self.retrieve()
           state.update(...)
           self.update(**state)

        where the original "id" field of self.retrieve() is harmlessly
        passed through as a keyword.

        """
        rep = self._digest_alias_args(id, owner, alias_target)
        if id is not None and id != self.alias_id:
            raise ValueError('parameter id must be None or %r, not %r' % (self.alias_id, id))
        return self.put('', json=rep).json()

    def delete_ermrest_alias(self, really=False):
        """Perform DELETE request, destroying alias on server.

        :param really: delete when True, abort when False (default)

        """
        if really is True:
            return DerivaBinding.delete(self, '')
        else:
            raise ValueError('Alias deletion refused when really is %s.' % really)




__version__ = "1.7.0"

from deriva.core.utils.core_utils import *
from deriva.core.base_cli import BaseCLI, KeyValuePairArgs
from deriva.core.deriva_binding import DerivaBinding, DerivaPathError, DerivaClientContext
from deriva.core.deriva_server import DerivaServer
from deriva.core.ermrest_catalog import ErmrestCatalog, ErmrestSnapshot, ErmrestCatalogMutationError, ErmrestAlias
from deriva.core.polling_ermrest_catalog import PollingErmrestCatalog
from deriva.core.hatrac_store import HatracStore, HatracHashMismatch, HatracJobPaused, HatracJobAborted, \
    HatracJobTimeout
from deriva.core.utils.globus_auth_utils import GlobusNativeLogin


def get_credential(host,
                   credential_file=DEFAULT_CREDENTIAL_FILE,
                   globus_credential_file=DEFAULT_GLOBUS_CREDENTIAL_FILE,
                   config_file=DEFAULT_CONFIG_FILE,
                   requested_scope=None,
                   force_scope_lookup=False,
                   match_scope_tag="deriva-all",
                   update_bdbag_keychain=True):
    """
    This function is used to get authorization credentials (in dict form) for use with various deriva-py API calls
    which take it as a parameter. A user must have already authenticated to the target host using either `deriva-auth`
    or `deriva-globus-auth-utils login` prior to calling this function, or the credential set for the host will not be
    found.

    :param host: The hostname to retrieve the credential set for.
    :param credential_file: Optional path to non-default location of the webauthn cookie credential file.
    :param globus_credential_file: Optional path to non-default location of the GlobusAuth bearer token store.
    :param config_file: Optional path to the non-default location of the deriva-py config file.
    :param requested_scope: Optional, specific scope request string for the given host. If not specified, the webauthn
        service on the host will be queried to determine the host-to-scope mapping that should be used.
    :param force_scope_lookup: Optional parameter to force the webauthn scope query and update the cached value in the
        deriva-py config file. A scope lookup will always be performed the first time a host-to-scope mapping is needed
        and is not already present in the configuration file for a given host.
    :param match_scope_tag: In the case that a host-to-scope mapping request returns multiple scopes, this is the key
        value ("tag") to match against in the result dict. By convention, the default is set to "deriva-all", which is
        the expected response from webauthn.
    :param update_bdbag_keychain: Updates the bdbag keychain file with the bearer token mapped to `host`. This is
        done to ensure that the bdbag keychain is updated when a refreshable bearer-token gets refreshed during the
        login check. Defaults to True.
    :return: A dict containing credential authorization values mapped by authorization type
    """
    # load deriva credential set first
    credentials = read_credential(credential_file or DEFAULT_CREDENTIAL_FILE, create_default=True)
    creds = credentials.get(host, credentials.get(host.lower(), dict()))

    # load globus credentials and merge, if present
    if globus_credential_file is not None and \
            os.path.isfile(globus_credential_file) and \
            os.path.getsize(globus_credential_file) > 10:  # Don't load empty json
        try:
            globus_client = GlobusNativeLogin(hosts=[host], config_file=config_file)
            scope_map = globus_client.hosts_to_scope_map(hosts=[host], match_scope_tag=match_scope_tag,
                                                         force_refresh=force_scope_lookup,
                                                         warn_on_discovery_failure=True if not creds else False)
            tokens = globus_client.is_logged_in(exclude_defaults=True)
            if tokens:
                # 1. look for the explicitly requested scope in the token store, if specified
                token = globus_client.find_access_token_for_scope(requested_scope, tokens)
                if not token:
                    # 2. try to determine the scope to use based on host-to-scope(s) mappings
                    token = globus_client.find_access_token_for_host(host,
                                                                     scope_map,
                                                                     tokens,
                                                                     match_scope_tag=match_scope_tag)
                if token:
                    creds["bearer-token"] = token
                    if update_bdbag_keychain:
                        globus_client.update_bdbag_keychain(token=token, host=host)
        except Exception as e:
            logging.warning("Exception while getting Globus credentials: %s" % format_exception(e))

    return creds or None

